# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Tuple

import wrapt
from opentelemetry.semconv.trace import SpanAttributes

from instana.log import logger
from instana.propagators.format import Format
from instana.singletons import agent, tracer
from instana.util.secrets import strip_secrets_from_query
from instana.util.traceutils import extract_custom_headers

if TYPE_CHECKING:
    from instana.span.span import InstanaSpan

try:
    import aiohttp
    from aiohttp.web import middleware

    if TYPE_CHECKING:
        import aiohttp.web

    @middleware
    async def stan_middleware(
        request: "aiohttp.web.Request",
        handler: Callable[..., object],
    ) -> Awaitable["aiohttp.web.Response"]:
        try:
            span_context = tracer.extract(Format.HTTP_HEADERS, request.headers)
            span: "InstanaSpan" = tracer.start_span(
                "aiohttp-server", span_context=span_context
            )
            request["span"] = span

            # Query param scrubbing
            url = str(request.url)
            parts = url.split("?")
            if len(parts) > 1:
                cleaned_qp = strip_secrets_from_query(
                    parts[1], agent.options.secrets_matcher, agent.options.secrets_list
                )
                span.set_attribute("http.params", cleaned_qp)

            span.set_attribute(SpanAttributes.HTTP_URL, parts[0])
            span.set_attribute(SpanAttributes.HTTP_METHOD, request.method)

            extract_custom_headers(span, request.headers)

            response = None
            try:
                response = await handler(request)
            except aiohttp.web.HTTPException as exc:
                # AIOHTTP uses exceptions for specific responses
                # see https://docs.aiohttp.org/en/latest/web_exceptions.html#web-server-exceptions
                response = exc

            if response is not None:
                # Mark 500 responses as errored
                if 500 <= response.status:
                    span.mark_as_errored()

                span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, response.status)

                extract_custom_headers(span, response.headers)

                tracer.inject(span.context, Format.HTTP_HEADERS, response.headers)

            return response
        except Exception as exc:
            logger.debug("aiohttp server stan_middleware:", exc_info=True)
            if span:
                span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, 500)
                span.record_exception(exc)
            raise
        finally:
            if span and span.is_recording():
                span.end()

    @wrapt.patch_function_wrapper("aiohttp.web", "Application.__init__")
    def init_with_instana(
        wrapped: Callable[..., "aiohttp.web.Application.__init__"],
        instance: "aiohttp.web.Application",
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> object:
        if "middlewares" in kwargs:
            kwargs["middlewares"].insert(0, stan_middleware)
        else:
            kwargs["middlewares"] = [stan_middleware]

        return wrapped(*args, **kwargs)

    logger.debug("Instrumenting aiohttp server")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instrumentation for AWS Lambda functions
"""

import sys
import traceback
from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple

import wrapt
from opentelemetry.semconv.trace import SpanAttributes

from instana import get_aws_lambda_handler
from instana.instrumentation.aws.triggers import enrich_lambda_span, get_context
from instana.log import logger
from instana.singletons import env_is_aws_lambda, get_agent, get_tracer
from instana.util.ids import define_server_timing

if TYPE_CHECKING:
    from instana.agent.aws_lambda import AWSLambdaAgent


def lambda_handler_with_instana(
    wrapped: Callable[..., object],
    instance: object,
    args: Tuple[object, ...],
    kwargs: Dict[str, Any],
) -> object:
    event = args[0]
    agent: "AWSLambdaAgent" = get_agent()
    tracer = get_tracer()

    agent.collector.collect_snapshot(*args)
    incoming_ctx = get_context(tracer, event)

    result = None
    with tracer.start_as_current_span(
        "aws.lambda.entry", span_context=incoming_ctx
    ) as span:
        enrich_lambda_span(agent, span, *args)
        try:
            result = wrapped(*args, **kwargs)

            if isinstance(result, dict):
                server_timing_value = define_server_timing(span.context.trace_id)
                if "headers" in result:
                    result["headers"]["Server-Timing"] = server_timing_value
                elif "multiValueHeaders" in result:
                    result["multiValueHeaders"]["Server-Timing"] = [server_timing_value]
                if "statusCode" in result and result.get("statusCode"):
                    status_code = int(result["statusCode"])
                    span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, status_code)
                    if 500 <= status_code:
                        span.record_exception(f"HTTP status {status_code}")
        except Exception as exc:
            logger.debug(f"AWS Lambda lambda_handler_with_instana error: {exc}")
            if span:
                exc = traceback.format_exc()
                span.record_exception(exc)
            raise
        finally:
            agent.collector.shutdown()

    if agent.collector.started:
        agent.collector.shutdown()
    
    return result


if env_is_aws_lambda:
    handler_module, handler_function = get_aws_lambda_handler()

    if handler_module and handler_function:
        try:
            logger.debug(
                f"Instrumenting AWS Lambda handler ({handler_module}.{handler_function})"
            )
            sys.path.insert(0, "/var/runtime")
            sys.path.insert(0, "/var/task")
            wrapt.wrap_function_wrapper(
                handler_module, handler_function, lambda_handler_with_instana
            )
        except (ModuleNotFoundError, ImportError) as exc:
            logger.debug(f"AWS Lambda error: {exc}")
            logger.warning(
                "Instana: Couldn't instrument AWS Lambda handler. Not monitoring."
            )
    else:
        logger.warning(
            "Instana: Couldn't determine AWS Lambda Handler.  Not monitoring."
        )



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


try:
    import tornado

    import wrapt

    from opentelemetry.semconv.trace import SpanAttributes

    from instana.log import logger
    from instana.singletons import agent, tracer
    from instana.util.secrets import strip_secrets_from_query
    from instana.util.traceutils import extract_custom_headers
    from instana.propagators.format import Format



    @wrapt.patch_function_wrapper('tornado.web', 'RequestHandler._execute')
    def execute_with_instana(wrapped, instance, argv, kwargs):
        try:
            span_context = None
            if hasattr(instance.request.headers, '__dict__') and '_dict' in instance.request.headers.__dict__:
                span_context = tracer.extract(Format.HTTP_HEADERS,
                                                instance.request.headers.__dict__['_dict'])

            span = tracer.start_span("tornado-server", span_context=span_context)

            # Query param scrubbing
            if instance.request.query is not None and len(instance.request.query) > 0:
                cleaned_qp = strip_secrets_from_query(instance.request.query, agent.options.secrets_matcher,
                                                        agent.options.secrets_list)
                span.set_attribute("http.params", cleaned_qp)
            
            url = f"{instance.request.protocol}://{instance.request.host}{instance.request.path}"
            span.set_attribute(SpanAttributes.HTTP_URL, url)
            span.set_attribute(SpanAttributes.HTTP_METHOD, instance.request.method)

            span.set_attribute("handler", instance.__class__.__name__)

            # Request header tracking support
            extract_custom_headers(span, instance.request.headers)

            setattr(instance.request, "_instana", span)

            # Set the context response headers now because tornado doesn't give us a better option to do so
            # later for this request.
            tracer.inject(span.context, Format.HTTP_HEADERS, instance._headers)

            return wrapped(*argv, **kwargs)
        except Exception:
            logger.debug("tornado execute", exc_info=True)


    @wrapt.patch_function_wrapper('tornado.web', 'RequestHandler.set_default_headers')
    def set_default_headers_with_instana(wrapped, instance, argv, kwargs):
        if not hasattr(instance.request, '_instana'):
            return wrapped(*argv, **kwargs)

        span = instance.request._instana
        tracer.inject(span.context, Format.HTTP_HEADERS, instance._headers)


    @wrapt.patch_function_wrapper('tornado.web', 'RequestHandler.on_finish')
    def on_finish_with_instana(wrapped, instance, argv, kwargs):
        try:
            if not hasattr(instance.request, '_instana'):
                return wrapped(*argv, **kwargs)

            span = instance.request._instana
            # Response header tracking support
            extract_custom_headers(span, instance._headers)

            status_code = instance.get_status()

            # Mark 500 responses as errored
            if 500 <= status_code:
                span.mark_as_errored()

            span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, status_code)
            if span.is_recording():
                span.end()

            return wrapped(*argv, **kwargs)
        except Exception:
            logger.debug("tornado on_finish", exc_info=True)


    @wrapt.patch_function_wrapper('tornado.web', 'RequestHandler.log_exception')
    def log_exception_with_instana(wrapped, instance, argv, kwargs):
        try:
            if not hasattr(instance.request, '_instana'):
                return wrapped(*argv, **kwargs)

            if not isinstance(argv[1], tornado.web.HTTPError):
                span = instance.request._instana
                span.record_exception(argv[0])

            return wrapped(*argv, **kwargs)
        except Exception:
            logger.debug("tornado log_exception", exc_info=True)


    logger.debug("Instrumenting tornado server")
except ImportError:
    pass



import logging
from typing import Generator
from unittest.mock import patch

import psycopg2
import psycopg2.extras
import pytest
from instana.instrumentation.pep0249 import (
    ConnectionFactory,
    ConnectionWrapper,
    CursorWrapper,
)
from instana.singletons import tracer
from instana.span.span import InstanaSpan
from opentelemetry.trace import SpanKind
from pytest import LogCaptureFixture

from tests.helpers import testenv


class TestCursorWrapper:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.connect_params = [
            "db",
            {
                "db": testenv["postgresql_db"],
                "host": testenv["postgresql_host"],
                "port": testenv["postgresql_port"],
                "user": testenv["postgresql_user"],
                "password": testenv["postgresql_pw"],
            },
        ]
        self.test_conn = psycopg2.connect(
            database=self.connect_params[1]["db"],
            host=self.connect_params[1]["host"],
            port=self.connect_params[1]["port"],
            user=self.connect_params[1]["user"],
            password=self.connect_params[1]["password"],
        )
        self.cursor_params = {"key": "value"}
        self.test_cursor = self.test_conn.cursor()
        self.cursor_name = "test-cursor"
        self.test_wrapper = CursorWrapper(
            self.test_cursor,
            self.cursor_name,
            self.connect_params,
            self.cursor_params,
        )
        yield
        self.test_cursor.close()
        self.test_conn.close()

    def reset_table(self) -> None:
        self.test_cursor.execute(
            """
            DROP TABLE IF EXISTS tests;
            CREATE TABLE tests (id SERIAL PRIMARY KEY, name VARCHAR(50), email VARCHAR(100));
            """
        )
        self.test_cursor.execute(
            """
            INSERT INTO tests (id, name, email) VALUES (1, 'test-name', 'testemail@mail.com');
            """
        )
        self.test_conn.commit()

    def reset_procedure(self) -> None:
        self.test_cursor.execute("""
            DROP PROCEDURE IF EXISTS insert_user(IN test_id INT, IN test_name VARCHAR, IN test_email VARCHAR);
            CREATE PROCEDURE insert_user(IN test_id INT, IN test_name VARCHAR, IN test_email VARCHAR)
            LANGUAGE plpgsql
            AS $$
            BEGIN
                INSERT INTO tests (id, name, email) VALUES (test_id, test_name, test_email);
            END;
            $$;
            """)
        self.test_conn.commit()

    def test_cursor_wrapper_default(self) -> None:
        # CursorWrapper
        assert self.test_wrapper
        assert self.test_wrapper._module_name == self.cursor_name
        connection_params = {"db", "host", "port", "user", "password"}
        assert connection_params.issubset(self.test_wrapper._connect_params[1].keys())
        assert not self.test_wrapper.closed
        assert self.test_wrapper._cursor_params == self.cursor_params

        # Test Connection
        assert (
            self.test_conn.dsn
            == "user=root password=xxx dbname=instana_test_db host=127.0.0.1 port=5432"
        )
        assert not self.test_conn.autocommit
        assert self.test_conn.status == 1
        assert self.test_conn.info.dbname == "instana_test_db"
        assert self.test_conn.info.host == "127.0.0.1"
        assert self.test_conn.info.user == "root"
        assert self.test_conn.info.port == 5432

        # Test Cursor
        assert self.test_cursor.arraysize == 1
        assert isinstance(self.test_cursor, psycopg2.extensions.cursor)
        assert hasattr(self.test_cursor, "callproc")
        assert hasattr(self.test_cursor, "close")
        assert hasattr(self.test_cursor, "execute")
        assert hasattr(self.test_cursor, "executemany")
        assert hasattr(self.test_cursor, "fetchone")
        assert hasattr(self.test_cursor, "fetchall")

    def test_collect_kvs(self) -> None:
        self.reset_table()
        with tracer.start_as_current_span("test") as span:
            sample_sql = """
                select * from tests;
            """
            self.test_wrapper._collect_kvs(span, sample_sql)
            assert span.attributes["span.kind"] == SpanKind.CLIENT
            assert span.attributes["db.name"] == "instana_test_db"
            assert span.attributes["db.statement"] == sample_sql
            assert span.attributes["db.user"] == "root"
            assert span.attributes["host"] == "127.0.0.1"
            assert span.attributes["port"] == 5432

    def test_collect_kvs_error(self, caplog: LogCaptureFixture) -> None:
        self.reset_table()
        with tracer.start_as_current_span("test") as span:
            connect_params = "sample"
            sample_wrapper = CursorWrapper(
                self.test_cursor,
                self.cursor_name,
                connect_params,
            )
            sample_sql = "select * from tests;"
            caplog.set_level(logging.DEBUG, logger="instana")
            sample_wrapper._collect_kvs(span, sample_sql)
            assert "string indices must be integers" in caplog.messages[0]

    def test_enter(self) -> None:
        response = self.test_wrapper.__enter__()
        assert response == self.test_wrapper
        assert isinstance(response, CursorWrapper)

    def test_execute_with_tracing_off(self) -> None:
        self.reset_table()
        with tracer.start_as_current_span("sqlalchemy"):
            sample_sql = """insert into tests (id, name, email) values (%s, %s, %s) returning id, name, email;"""
            sample_params = (2, "sample-name", "sample-email@mail.com")
            self.test_wrapper.execute(sample_sql, sample_params)
            self.test_wrapper.execute("select * from tests;")
            response = self.test_wrapper.fetchall()
            assert sample_params in response
            assert len(response) == 2

    def test_execute_with_tracing(self) -> None:
        self.reset_table()
        with tracer.start_as_current_span("test"):
            sample_sql = """insert into tests (id, name, email) values (%s, %s, %s) returning id, name, email;"""
            sample_params = (3, "sample-name", "sample-email@mail.com")
            self.test_wrapper.execute(sample_sql, sample_params)
            last_inserted_row = self.test_cursor.fetchone()
            self.test_conn.commit()
            assert last_inserted_row == sample_params

            # Exception Handling
            with pytest.raises(Exception) as exc_info, patch.object(
                CursorWrapper, "_collect_kvs", side_effect=Exception("test exception")
            ) as mock_collect_kvs:
                self.test_wrapper.execute(sample_sql)
            assert str(exc_info.value) == "test exception"
            mock_collect_kvs.assert_called_once()
            self.test_wrapper.execute("select * from tests;")
            response = self.test_wrapper.fetchall()
            assert sample_params in response
            assert len(response) == 2

    def test_executemany_with_tracing_off(self) -> None:
        self.reset_table()
        with tracer.start_as_current_span("sqlalchemy"):
            sample_sql = """insert into tests (id, name, email) values (%s, %s, %s) returning id, name, email;"""
            sample_seq_of_params = [
                (4, "sample-name-3", "sample-email-3@mail.com"),
                (5, "sample-name-4", "sample-email-4@mail.com"),
            ]
            self.test_wrapper.executemany(sample_sql, sample_seq_of_params)
            self.test_wrapper.execute("select * from tests;")
            response = self.test_wrapper.fetchall()
            for record in sample_seq_of_params:
                assert record in response
            assert len(response) == 3

    def test_executemany_with_tracing(self) -> None:
        self.reset_table()
        with tracer.start_as_current_span("test"):
            sample_sql = """insert into tests (id, name, email) values (%s, %s, %s) returning id, name, email;"""
            sample_seq_of_params = [
                (6, "sample-name-3", "sample-email-3@mail.com"),
                (7, "sample-name-4", "sample-email-4@mail.com"),
            ]
            self.test_wrapper.executemany(sample_sql, sample_seq_of_params)

            # Exception Handling
            with pytest.raises(Exception) as exc_info, patch.object(
                CursorWrapper, "_collect_kvs", side_effect=Exception("test exception")
            ) as mock_collect_kvs:
                self.test_wrapper.executemany(
                    sample_sql, seq_of_parameters=sample_seq_of_params
                )
            assert str(exc_info.value) == "test exception"
            mock_collect_kvs.assert_called_once()
            self.test_wrapper.execute("select * from tests;")
            response = self.test_wrapper.fetchall()
            for record in sample_seq_of_params:
                assert record in response
            assert len(response) == 3

    def test_callproc_with_tracing_off(self) -> None:
        self.reset_table()
        self.reset_procedure()
        with tracer.start_as_current_span("sqlalchemy"):
            sample_proc_name = "call insert_user(%s, %s, %s);"
            sample_params = (8, "sample-name-8", "sample-email-8@mail.com")
            self.test_wrapper.callproc(sample_proc_name, sample_params)
            self.test_conn.commit()
            self.test_wrapper.execute("select * from tests;")
            response = self.test_wrapper.fetchall()
            assert sample_params in response
            assert len(response) == 2

    def test_callproc_with_tracing(self) -> None:
        self.reset_table()
        self.reset_procedure()
        with tracer.start_as_current_span("test"):
            sample_proc_name = "call insert_user(%s, %s, %s);"
            sample_params = (9, "sample-name-9", "sample-email-9@mail.com")
            self.test_wrapper.callproc(sample_proc_name, sample_params)
            self.test_conn.commit()
            self.test_wrapper.execute("select * from tests;")
            response = self.test_wrapper.fetchall()
            assert sample_params in response
            assert len(response) == 2

            # Exception Handling
            error_proc_name = "erroroeus command;"
            with pytest.raises(Exception) as exc_info, patch.object(
                InstanaSpan,
                "record_exception",
            ) as mock_exception:
                self.test_wrapper.callproc(error_proc_name, sample_params)
            assert exc_info.typename == "SyntaxError"
            mock_exception.call_count == 2


class TestConnectionWrapper:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.connect_params = [
            "db",
            {
                "db": "instana_test_db",
                "host": "localhost",
                "port": "5432",
                "user": "root",
                "password": "passw0rd",
            },
        ]
        self.test_conn = psycopg2.connect(
            database=self.connect_params[1]["db"],
            host=self.connect_params[1]["host"],
            port=self.connect_params[1]["port"],
            user=self.connect_params[1]["user"],
            password=self.connect_params[1]["password"],
        )
        self.module_name = "test-connection"
        self.connection_manager = ConnectionWrapper(
            self.test_conn, self.module_name, self.connect_params
        )
        yield
        self.test_conn.close()

    def test_enter(self) -> None:
        response = self.connection_manager.__enter__()
        assert isinstance(response, ConnectionWrapper)
        assert response._module_name == self.module_name
        assert response._connect_params == self.connect_params

    def test_cursor(self) -> None:
        response = self.connection_manager.cursor()
        assert isinstance(response, CursorWrapper)

    def test_close(self) -> None:
        response = self.connection_manager.close()
        assert self.test_conn.closed
        assert not response

    def test_commit(self) -> None:
        response = self.connection_manager.commit()
        assert not response

    def test_rollback(self) -> None:
        if hasattr(self.connection_manager, "rollback"):
            response = self.connection_manager.rollback()
            assert not response


class TestConnectionFactory:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.test_conn_func = psycopg2.connect
        self.test_module_name = "test-factory"
        self.conn_fact = ConnectionFactory(self.test_conn_func, self.test_module_name)
        yield
        self.test_conn_func = None
        self.test_module_name = None
        self.conn_fact = None

    def test_call(self) -> None:
        response = self.conn_fact(
            dsn="user=root password=passw0rd dbname=instana_test_db host=localhost port=5432"
        )
        assert isinstance(self.conn_fact._wrapper_ctor, ConnectionWrapper.__class__)
        assert self.conn_fact._connect_func == self.test_conn_func
        assert self.conn_fact._module_name == self.test_module_name
        assert isinstance(response, ConnectionWrapper)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import logging
import sys
from multiprocessing.pool import ThreadPool
from time import sleep
from typing import TYPE_CHECKING, Generator

import pytest
import requests
import urllib3
from instana.instrumentation.urllib3 import (
    _collect_kvs as collect_kvs,
    extract_custom_headers,
    collect_response,
)
from instana.singletons import agent, tracer

import tests.apps.flask_app  # noqa: F401
from tests.helpers import testenv

if TYPE_CHECKING:
    from instana.span.span import InstanaSpan
    from pytest import LogCaptureFixture


class TestUrllib3:
    @pytest.fixture(autouse=True)
    def _setup(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.http = urllib3.PoolManager()
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        yield
        # teardown
        # Ensure that allow_exit_as_root has the default value"""
        agent.options.allow_exit_as_root = False

    def test_vanilla_requests(self) -> None:
        r = self.http.request("GET", testenv["flask_server"] + "/")
        assert r.status == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

    def test_parallel_requests(self) -> None:
        http_pool_5 = urllib3.PoolManager(num_pools=5)

        def task(num):
            r = http_pool_5.request(
                "GET", testenv["flask_server"] + "/", fields={"num": num}
            )
            return r

        with ThreadPool(processes=5) as executor:
            # iterate over results as they become available
            for result in executor.map(task, (1, 2, 3, 4, 5)):
                assert result.status == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 5
        nums = map(lambda s: s.data["http"]["params"].split("=")[1], spans)
        assert set(nums) == set(("1", "2", "3", "4", "5"))

    @pytest.mark.skipif(
        sys.platform == "darwin",
        reason="Avoiding ConnectionError when calling multi processes of Flask app.",
    )
    def test_customers_setup_zd_26466(self) -> None:
        def make_request(u=None) -> int:
            sleep(10)
            x = requests.get(testenv["flask_server"] + "/")
            sleep(10)
            return x.status_code

        status = make_request()
        assert status == 200
        # print(f'request made outside threadpool, instana should instrument - status: {status}')

        threadpool_size = 15
        pool = ThreadPool(processes=threadpool_size)
        res = pool.map(make_request, [u for u in range(threadpool_size)])
        # print(f'requests made within threadpool, instana does not instrument - statuses: {res}')

        spans = self.recorder.queued_spans()
        assert len(spans) == 16

    def test_get_request(self):
        with tracer.start_as_current_span("test"):
            r = self.http.request("GET", testenv["flask_server"] + "/")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status == 200

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 200
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_get_request_https(self):
        request_url = "https://reqres.in:443/api/users"
        with tracer.start_as_current_span("test"):
            r = self.http.request("GET", request_url)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        urllib3_span = spans[0]
        test_span = spans[1]

        assert r
        assert r.status == 200

        # Same traceId
        assert test_span.t == urllib3_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert urllib3_span.data["http"]["url"] == request_url
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_get_request_as_root_exit_span(self):
        agent.options.allow_exit_as_root = True
        r = self.http.request("GET", testenv["flask_server"] + "/")

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        wsgi_span = spans[0]
        urllib3_span = spans[1]

        assert r
        assert r.status == 200
        # assert not tracer.active_span

        # Same traceId
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert not urllib3_span.p
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 200
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_get_request_with_query(self):
        with tracer.start_as_current_span("test"):
            r = self.http.request("GET", testenv["flask_server"] + "/?one=1&two=2")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status == 200
        # assert not tracer.active_span

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 200
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span.data["http"]["params"] in ["one=1&two=2", "two=2&one=1"]
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_get_request_with_alt_query(self):
        with tracer.start_as_current_span("test"):
            r = self.http.request(
                "GET", testenv["flask_server"] + "/", fields={"one": "1", "two": 2}
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status == 200
        # assert not tracer.active_span

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 200
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span.data["http"]["params"] in ["one=1&two=2", "two=2&one=1"]
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_put_request(self):
        with tracer.start_as_current_span("test"):
            r = self.http.request("PUT", testenv["flask_server"] + "/notfound")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status == 404
        # assert not tracer.active_span

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/notfound"
        assert wsgi_span.data["http"]["method"] == "PUT"
        assert wsgi_span.data["http"]["status"] == 404
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 404
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/notfound"
        assert urllib3_span.data["http"]["method"] == "PUT"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_301_redirect(self):
        with tracer.start_as_current_span("test"):
            r = self.http.request("GET", testenv["flask_server"] + "/301")

        spans = self.recorder.queued_spans()
        assert len(spans) == 5

        wsgi_span2 = spans[0]
        urllib3_span2 = spans[1]
        wsgi_span1 = spans[2]
        urllib3_span1 = spans[3]
        test_span = spans[4]

        assert r
        assert r.status == 200
        # assert not tracer.active_span

        # Same traceId
        traceId = test_span.t
        assert urllib3_span1.t == traceId
        assert wsgi_span1.t == traceId
        assert urllib3_span2.t == traceId
        assert wsgi_span2.t == traceId

        # Parent relationships
        assert urllib3_span1.p == test_span.s
        assert wsgi_span1.p == urllib3_span1.s
        assert urllib3_span2.p == test_span.s
        assert wsgi_span2.p == urllib3_span2.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span1.ec
        assert not wsgi_span1.ec
        assert not urllib3_span2.ec
        assert not wsgi_span2.ec

        # wsgi
        assert wsgi_span1.n == "wsgi"
        assert wsgi_span1.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span1.data["http"]["url"] == "/"
        assert wsgi_span1.data["http"]["method"] == "GET"
        assert wsgi_span1.data["http"]["status"] == 200
        assert not wsgi_span1.data["http"]["error"]
        assert not wsgi_span1.stack

        assert wsgi_span2.n == "wsgi"
        assert wsgi_span2.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span2.data["http"]["url"] == "/301"
        assert wsgi_span2.data["http"]["method"] == "GET"
        assert wsgi_span2.data["http"]["status"] == 301
        assert not wsgi_span2.data["http"]["error"]
        assert not wsgi_span2.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span1.n == "urllib3"
        assert urllib3_span1.data["http"]["status"] == 200
        assert urllib3_span1.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span1.data["http"]["method"] == "GET"
        assert urllib3_span1.stack
        assert isinstance(urllib3_span1.stack, list)
        assert len(urllib3_span1.stack) > 1

        assert urllib3_span2.n == "urllib3"
        assert urllib3_span2.data["http"]["status"] == 301
        assert urllib3_span2.data["http"]["url"] == testenv["flask_server"] + "/301"
        assert urllib3_span2.data["http"]["method"] == "GET"
        assert urllib3_span2.stack
        assert isinstance(urllib3_span2.stack, list)
        assert len(urllib3_span2.stack) > 1

    def test_302_redirect(self):
        with tracer.start_as_current_span("test"):
            r = self.http.request("GET", testenv["flask_server"] + "/302")

        spans = self.recorder.queued_spans()
        assert len(spans) == 5

        wsgi_span2 = spans[0]
        urllib3_span2 = spans[1]
        wsgi_span1 = spans[2]
        urllib3_span1 = spans[3]
        test_span = spans[4]

        assert r
        assert r.status == 200
        # assert not tracer.active_span

        # Same traceId
        traceId = test_span.t
        assert urllib3_span1.t == traceId
        assert wsgi_span1.t == traceId
        assert urllib3_span2.t == traceId
        assert wsgi_span2.t == traceId

        # Parent relationships
        assert urllib3_span1.p == test_span.s
        assert wsgi_span1.p == urllib3_span1.s
        assert urllib3_span2.p == test_span.s
        assert wsgi_span2.p == urllib3_span2.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span1.ec
        assert not wsgi_span1.ec
        assert not urllib3_span2.ec
        assert not wsgi_span2.ec

        # wsgi
        assert wsgi_span1.n == "wsgi"
        assert wsgi_span1.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span1.data["http"]["url"] == "/"
        assert wsgi_span1.data["http"]["method"] == "GET"
        assert wsgi_span1.data["http"]["status"] == 200
        assert not wsgi_span1.data["http"]["error"]
        assert not wsgi_span1.stack

        assert wsgi_span2.n == "wsgi"
        assert wsgi_span2.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span2.data["http"]["url"] == "/302"
        assert wsgi_span2.data["http"]["method"] == "GET"
        assert wsgi_span2.data["http"]["status"] == 302
        assert not wsgi_span2.data["http"]["error"]
        assert not wsgi_span2.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span1.n == "urllib3"
        assert urllib3_span1.data["http"]["status"] == 200
        assert urllib3_span1.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span1.data["http"]["method"] == "GET"
        assert urllib3_span1.stack
        assert isinstance(urllib3_span1.stack, list)
        assert len(urllib3_span1.stack) > 1

        assert urllib3_span2.n == "urllib3"
        assert urllib3_span2.data["http"]["status"] == 302
        assert urllib3_span2.data["http"]["url"] == testenv["flask_server"] + "/302"
        assert urllib3_span2.data["http"]["method"] == "GET"
        assert urllib3_span2.stack
        assert isinstance(urllib3_span2.stack, list)
        assert len(urllib3_span2.stack) > 1

    def test_5xx_request(self):
        with tracer.start_as_current_span("test"):
            r = self.http.request("GET", testenv["flask_server"] + "/504")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status == 504
        # assert not tracer.active_span

        # Same traceId
        traceId = test_span.t
        assert urllib3_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert urllib3_span.ec == 1
        assert wsgi_span.ec == 1

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/504"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 504
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 504
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/504"
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_exception_logging(self):
        with tracer.start_as_current_span("test"):
            try:
                r = self.http.request("GET", testenv["flask_server"] + "/exception")
            except Exception:
                pass

        spans = self.recorder.queued_spans()
        # Behind the "wsgi_server", currently there is Flask
        # Flask <  2.3.0 optionally can depend on "blinker"
        # Flask >= 2.3.0 unconditionally depends on "blinker"
        # Depending on whether we instrument with "flask/vanilla.py" or "flask/with_blinker.py"
        # The exception logging differs. See the log_exception_with_instana function in flask/with_blinker.py
        # which is called in the blinker scenario.
        # Without blinker, Flask does some extra logging, which results an extra log span recorded
        # but was disregarded by this TC anyway, so for the rest of the TC
        # we will just discard the optional log span if present
        # Without blinker, our instrumentation logs roughly the same exception data onto the
        # already existing wsgi span. Which we validate in this TC if present.
        assert len(spans) in (3, 4)

        with_blinker = len(spans) == 3
        if not with_blinker:
            spans = spans[1:]

        wsgi_span, urllib3_span, test_span = spans

        assert r
        assert r.status == 500
        # assert not tracer.active_span

        # Same traceId
        traceId = test_span.t
        assert urllib3_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert urllib3_span.ec == 1
        assert wsgi_span.ec == 1

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/exception"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 500
        if with_blinker:
            assert wsgi_span.data["http"]["error"] == "fake error"
        else:
            assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 500
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/exception"
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_client_error(self):
        r = None
        with tracer.start_as_current_span("test"):
            try:
                r = self.http.request(
                    "GET",
                    "http://doesnotexist.asdf:5000/504",
                    retries=False,
                    timeout=urllib3.Timeout(connect=0.5, read=0.5),
                )
            except Exception:
                pass

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        urllib3_span = spans[0]
        test_span = spans[1]

        assert not r

        # Parent relationships
        assert urllib3_span.p == test_span.s

        # Same traceId
        traceId = test_span.t
        assert urllib3_span.t == traceId

        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert not urllib3_span.data["http"]["status"]
        assert urllib3_span.data["http"]["url"] == "http://doesnotexist.asdf:5000/504"
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

        # Error logging
        assert not test_span.ec
        assert urllib3_span.ec == 2

    def test_requests_pkg_get(self):
        self.recorder.clear_spans()

        with tracer.start_as_current_span("test"):
            r = requests.get(testenv["flask_server"] + "/", timeout=2)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status_code == 200
        # assert not tracer.active_span

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 200
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_requests_pkg_get_with_custom_headers(self):
        my_custom_headers = dict()
        my_custom_headers["X-PGL-1"] = "1"

        with tracer.start_as_current_span("test"):
            r = requests.get(
                testenv["flask_server"] + "/", timeout=2, headers=my_custom_headers
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status_code == 200
        # assert not tracer.active_span

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 200
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_requests_pkg_put(self):
        with tracer.start_as_current_span("test"):
            r = requests.put(testenv["flask_server"] + "/notfound")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r.status_code == 404
        # assert not tracer.active_span

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/notfound"
        assert wsgi_span.data["http"]["method"] == "PUT"
        assert wsgi_span.data["http"]["status"] == 404
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 404
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/notfound"
        assert urllib3_span.data["http"]["method"] == "PUT"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

    def test_response_header_capture(self):
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        with tracer.start_as_current_span("test"):
            r = self.http.request("GET", testenv["flask_server"] + "/response_headers")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status == 200
        # assert not tracer.active_span

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/response_headers"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 200
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert (
            urllib3_span.data["http"]["url"]
            == testenv["flask_server"] + "/response_headers"
        )
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

        assert "X-Capture-This" in urllib3_span.data["http"]["header"]
        assert urllib3_span.data["http"]["header"]["X-Capture-This"] == "Ok"
        assert "X-Capture-That" in urllib3_span.data["http"]["header"]
        assert urllib3_span.data["http"]["header"]["X-Capture-That"] == "Ok too"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_request_header_capture(self):
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        request_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }
        with tracer.start_as_current_span("test"):
            r = self.http.request(
                "GET", testenv["flask_server"] + "/", headers=request_headers
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert r
        assert r.status == 200
        # assert not tracer.active_span

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert wsgi_span.data["http"]["url"] == "/"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == 200
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert urllib3_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert isinstance(urllib3_span.stack, list)
        assert len(urllib3_span.stack) > 1

        assert "X-Capture-This-Too" in urllib3_span.data["http"]["header"]
        assert urllib3_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in urllib3_span.data["http"]["header"]
        assert urllib3_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_extract_custom_headers_exception(
        self, span: "InstanaSpan", caplog: "LogCaptureFixture", monkeypatch
    ) -> None:
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        request_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }

        monkeypatch.setattr(span, "set_attribute", Exception("mocked error"))
        caplog.set_level(logging.DEBUG, logger="instana")
        extract_custom_headers(span, request_headers)
        assert "extract_custom_headers: " in caplog.messages

    def test_collect_response_exception(
        self, span: "InstanaSpan", caplog: "LogCaptureFixture", monkeypatch
    ) -> None:
        monkeypatch.setattr(span, "set_attribute", Exception("mocked error"))

        caplog.set_level(logging.DEBUG, logger="instana")
        collect_response(span, {})
        assert "urllib3 collect_response error: " in caplog.messages

    def test_collect_kvs_exception(
        self, span: "InstanaSpan", caplog: "LogCaptureFixture", monkeypatch
    ) -> None:
        monkeypatch.setattr(span, "set_attribute", Exception("mocked error"))

        caplog.set_level(logging.DEBUG, logger="instana")
        collect_kvs({}, (), {})
        assert "urllib3 _collect_kvs error: " in caplog.messages



# (c) Copyright IBM Corp. 2024

"""Module to handle the collection of containerized process metrics for EKS Pods on AWS Fargate"""

import os

from instana.collector.helpers.process import ProcessHelper
from instana.log import logger


def get_pod_name():
    podname = os.environ.get("HOSTNAME", "")

    if not podname:
        logger.warning("Failed to determine podname from EKS hostname.")
    return podname


class EKSFargateProcessHelper(ProcessHelper):
    """Helper class to extend the generic process helper class with the corresponding fargate attributes"""

    def collect_metrics(self, **kwargs):
        plugin_data = dict()
        try:
            plugin_data = super(EKSFargateProcessHelper, self).collect_metrics(**kwargs)
            plugin_data["data"]["containerType"] = "docker"

            if kwargs.get("with_snapshot"):
                plugin_data["data"]["com.instana.plugin.host.name"] = get_pod_name()
        except Exception:
            logger.debug("EKSFargateProcessHelper.collect_metrics: ", exc_info=True)
        return [plugin_data]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright IBM Corp. 2024

import asyncio

from aiohttp import web

from tests.helpers import testenv

testenv["aiohttp_port"] = 10810
testenv["aiohttp_server"] = f"http://127.0.0.1:{testenv['aiohttp_port']}"


def say_hello(request):
    return web.Response(text="Hello, world")


@web.middleware
async def middleware1(request, handler):
    print("Middleware 1 called")
    response = await handler(request)
    print("Middleware 1 finished")
    return response


def aiohttp_server():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    app = web.Application(middlewares=[middleware1])
    app.add_routes([web.get("/", say_hello)])

    runner = web.AppRunner(app)
    loop.run_until_complete(runner.setup())
    site = web.TCPSite(runner, "127.0.0.1", testenv["aiohttp_port"])

    loop.run_until_complete(site.start())
    loop.run_forever()



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
AWS Fargate Collector: Manages the periodic collection of metrics & snapshot data
"""

import json
import os
from time import time

import requests

from instana.collector.base import BaseCollector
from instana.collector.helpers.fargate.container import ContainerHelper
from instana.collector.helpers.fargate.docker import DockerHelper
from instana.collector.helpers.fargate.process import FargateProcessHelper
from instana.collector.helpers.fargate.task import TaskHelper
from instana.collector.helpers.runtime import RuntimeHelper
from instana.collector.utils import format_span
from instana.log import logger
from instana.util import DictionaryOfStan, validate_url


class AWSFargateCollector(BaseCollector):
    """Collector for AWS Fargate"""

    def __init__(self, agent):
        super(AWSFargateCollector, self).__init__(agent)
        logger.debug("Loading AWS Fargate Collector")

        # Indicates if this Collector has all requirements to run successfully
        self.ready_to_start = True

        # Prepare the URLS that we will collect data from
        self.ecmu = os.environ.get("ECS_CONTAINER_METADATA_URI", "")

        if self.ecmu == "" or validate_url(self.ecmu) is False:
            logger.warning(
                "AWSFargateCollector: ECS_CONTAINER_METADATA_URI not in environment or invalid URL.  "
                "Instana will not be able to monitor this environment"
            )
            self.ready_to_start = False

        self.ecmu_url_root = self.ecmu
        self.ecmu_url_task = self.ecmu + "/task"
        self.ecmu_url_stats = self.ecmu + "/stats"
        self.ecmu_url_task_stats = self.ecmu + "/task/stats"

        # Timestamp in seconds of the last time we fetched all ECMU data
        self.last_ecmu_full_fetch = 0

        # How often to do a full fetch of ECMU data
        self.ecmu_full_fetch_interval = 304

        # HTTP client with keep-alive
        self.http_client = requests.Session()

        # This is the collecter thread querying the metadata url
        self.ecs_metadata_thread = None

        # The fully qualified ARN for this process
        self._fq_arn = None

        # Response from the last call to
        # ${ECS_CONTAINER_METADATA_URI}/
        self.root_metadata = None

        # Response from the last call to
        # ${ECS_CONTAINER_METADATA_URI}/task
        self.task_metadata = None

        # Response from the last call to
        # ${ECS_CONTAINER_METADATA_URI}/stats
        self.stats_metadata = None

        # Response from the last call to
        # ${ECS_CONTAINER_METADATA_URI}/task/stats
        self.task_stats_metadata = None

        # Populate the collection helpers
        self.helpers.append(TaskHelper(self))
        self.helpers.append(DockerHelper(self))
        self.helpers.append(FargateProcessHelper(self))
        self.helpers.append(RuntimeHelper(self))
        self.helpers.append(ContainerHelper(self))

    def start(self):
        if self.ready_to_start is False:
            logger.warning(
                "AWS Fargate Collector is missing requirements and cannot monitor this environment."
            )
            return

        super(AWSFargateCollector, self).start()

    def get_ecs_metadata(self):
        """
        Get the latest data from the ECS metadata container API and store on the class
        @return: Boolean
        """
        try:
            self.fetching_start_time = int(time())
            delta = self.fetching_start_time - self.last_ecmu_full_fetch
            if delta > self.ecmu_full_fetch_interval:
                # Refetch the ECMU snapshot data
                self.last_ecmu_full_fetch = int(time())

                # Response from the last call to
                # ${ECS_CONTAINER_METADATA_URI}/
                json_body = self.http_client.get(self.ecmu_url_root, timeout=1).content
                self.root_metadata = json.loads(json_body)

                # Response from the last call to
                # ${ECS_CONTAINER_METADATA_URI}/task
                json_body = self.http_client.get(self.ecmu_url_task, timeout=1).content
                self.task_metadata = json.loads(json_body)

            # Response from the last call to
            # ${ECS_CONTAINER_METADATA_URI}/stats
            json_body = self.http_client.get(self.ecmu_url_stats, timeout=2).content
            self.stats_metadata = json.loads(json_body)

            # Response from the last call to
            # ${ECS_CONTAINER_METADATA_URI}/task/stats
            json_body = self.http_client.get(
                self.ecmu_url_task_stats, timeout=1
            ).content
            self.task_stats_metadata = json.loads(json_body)
        except Exception:
            logger.debug("AWSFargateCollector.get_ecs_metadata", exc_info=True)

    def should_send_snapshot_data(self):
        return int(time()) - self.snapshot_data_last_sent > self.snapshot_data_interval

    def prepare_payload(self):
        payload = DictionaryOfStan()
        payload["spans"] = []
        payload["metrics"]["plugins"] = []

        try:
            if not self.span_queue.empty():
                payload["spans"] = format_span(self.queued_spans())

            with_snapshot = self.should_send_snapshot_data()

            # Fetch the latest metrics
            self.get_ecs_metadata()

            plugins = []
            for helper in self.helpers:
                plugins.extend(helper.collect_metrics(with_snapshot=with_snapshot))

            payload["metrics"]["plugins"] = plugins

            if with_snapshot is True:
                self.snapshot_data_last_sent = int(time())
        except Exception:
            logger.debug("collect_snapshot error", exc_info=True)

        return payload

    def get_fq_arn(self):
        if self._fq_arn is not None:
            return self._fq_arn

        task_arn = ""
        if self.root_metadata is not None:
            labels = self.root_metadata.get("Labels", None)
            if labels is not None:
                task_arn = labels.get("com.amazonaws.ecs.task-arn", "")

            container_name = self.root_metadata.get("Name", "")

            self._fq_arn = task_arn + "::" + container_name
            return self._fq_arn
        else:
            return "Missing ECMU metadata"



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

import logging
import os
from typing import Generator

import pytest

from instana.agent.google_cloud_run import GCRAgent
from instana.options import GCROptions
from instana.recorder import StanRecorder
from instana.singletons import get_agent, get_tracer, set_agent, set_tracer
from instana.tracer import InstanaTracer, InstanaTracerProvider


class TestGCR:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.agent = None
        self.span_recorder = None
        self.tracer = None

        self.original_agent = get_agent()
        self.original_tracer = get_tracer()

        os.environ["K_SERVICE"] = "service"
        os.environ["K_CONFIGURATION"] = "configuration"
        os.environ["K_REVISION"] = "revision"
        os.environ["PORT"] = "port"
        os.environ["INSTANA_ENDPOINT_URL"] = "https://localhost/notreal"
        os.environ["INSTANA_AGENT_KEY"] = "Fake_Key"
        yield
        if "K_SERVICE" in os.environ:
            os.environ.pop("K_SERVICE")
        if "K_CONFIGURATION" in os.environ:
            os.environ.pop("K_CONFIGURATION")
        if "K_REVISION" in os.environ:
            os.environ.pop("K_REVISION")
        if "PORT" in os.environ:
            os.environ.pop("PORT")
        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            os.environ.pop("INSTANA_EXTRA_HTTP_HEADERS")
        if "INSTANA_ENDPOINT_URL" in os.environ:
            os.environ.pop("INSTANA_ENDPOINT_URL")
        if "INSTANA_ENDPOINT_PROXY" in os.environ:
            os.environ.pop("INSTANA_ENDPOINT_PROXY")
        if "INSTANA_AGENT_KEY" in os.environ:
            os.environ.pop("INSTANA_AGENT_KEY")
        if "INSTANA_LOG_LEVEL" in os.environ:
            os.environ.pop("INSTANA_LOG_LEVEL")
        if "INSTANA_SECRETS" in os.environ:
            os.environ.pop("INSTANA_SECRETS")
        if "INSTANA_DEBUG" in os.environ:
            os.environ.pop("INSTANA_DEBUG")
        if "INSTANA_TAGS" in os.environ:
            os.environ.pop("INSTANA_TAGS")

        set_agent(self.original_agent)
        set_tracer(self.original_tracer)

    def create_agent_and_setup_tracer(
        self, tracer_provider: InstanaTracerProvider
    ) -> None:
        self.agent = GCRAgent(
            service="service",
            configuration="configuration",
            revision="revision",
        )
        self.span_processor = StanRecorder(self.agent)
        self.tracer = InstanaTracer(
            tracer_provider.sampler,
            self.span_processor,
            tracer_provider._exporter,
            tracer_provider._propagators,
        )
        set_agent(self.agent)
        set_tracer(self.tracer)

    def test_has_options(self, tracer_provider: InstanaTracerProvider) -> None:
        self.create_agent_and_setup_tracer(tracer_provider=tracer_provider)
        assert hasattr(self.agent, "options")
        assert isinstance(self.agent.options, GCROptions)

    def test_invalid_options(self):
        # None of the required env vars are available...
        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            os.environ.pop("INSTANA_EXTRA_HTTP_HEADERS")
        if "INSTANA_ENDPOINT_URL" in os.environ:
            os.environ.pop("INSTANA_ENDPOINT_URL")
        if "INSTANA_AGENT_KEY" in os.environ:
            os.environ.pop("INSTANA_AGENT_KEY")

        agent = GCRAgent(
            service="service", configuration="configuration", revision="revision"
        )
        assert not agent.can_send()
        assert not agent.collector

    def test_default_secrets(self, tracer_provider: InstanaTracerProvider) -> None:
        self.create_agent_and_setup_tracer(tracer_provider=tracer_provider)
        assert not self.agent.options.secrets
        assert hasattr(self.agent.options, "secrets_matcher")
        assert self.agent.options.secrets_matcher == "contains-ignore-case"
        assert hasattr(self.agent.options, "secrets_list")
        assert self.agent.options.secrets_list == ["key", "pass", "secret"]

    def test_custom_secrets(self, tracer_provider: InstanaTracerProvider) -> None:
        os.environ["INSTANA_SECRETS"] = "equals:love,war,games"
        self.create_agent_and_setup_tracer(tracer_provider=tracer_provider)

        assert hasattr(self.agent.options, "secrets_matcher")
        assert self.agent.options.secrets_matcher == "equals"
        assert hasattr(self.agent.options, "secrets_list")
        assert self.agent.options.secrets_list == ["love", "war", "games"]

    def test_has_extra_http_headers(
        self, tracer_provider: InstanaTracerProvider
    ) -> None:
        self.create_agent_and_setup_tracer(tracer_provider=tracer_provider)
        assert hasattr(self.agent, "options")
        assert hasattr(self.agent.options, "extra_http_headers")

    def test_agent_extra_http_headers(
        self, tracer_provider: InstanaTracerProvider
    ) -> None:
        os.environ["INSTANA_EXTRA_HTTP_HEADERS"] = (
            "X-Test-Header;X-Another-Header;X-And-Another-Header"
        )
        self.create_agent_and_setup_tracer(tracer_provider=tracer_provider)
        assert self.agent.options.extra_http_headers
        should_headers = ["x-test-header", "x-another-header", "x-and-another-header"]
        assert should_headers == self.agent.options.extra_http_headers

    def test_agent_default_log_level(
        self, tracer_provider: InstanaTracerProvider
    ) -> None:
        self.create_agent_and_setup_tracer(tracer_provider=tracer_provider)
        assert self.agent.options.log_level == logging.WARNING

    def test_agent_custom_log_level(
        self, tracer_provider: InstanaTracerProvider
    ) -> None:
        os.environ["INSTANA_LOG_LEVEL"] = "eRror"
        self.create_agent_and_setup_tracer(tracer_provider=tracer_provider)
        assert self.agent.options.log_level == logging.ERROR

    def test_custom_proxy(self, tracer_provider: InstanaTracerProvider) -> None:
        os.environ["INSTANA_ENDPOINT_PROXY"] = "http://myproxy.123"
        self.create_agent_and_setup_tracer(tracer_provider=tracer_provider)
        assert self.agent.options.endpoint_proxy == {"https": "http://myproxy.123"}



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import pytest
from typing import Generator

import asyncio
import aiohttp
import tornado
from tornado.httpclient import AsyncHTTPClient

from instana.util.ids import hex_id
import tests.apps.tornado_server

from instana.singletons import tracer, agent
from tests.helpers import testenv, get_first_span_by_name, get_first_span_by_filter
from instana.span.span import get_current_span


class TestTornadoServer:
    async def fetch(self, session, url, headers=None, params=None):
        try:
            async with session.get(url, headers=headers, params=params) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    async def post(self, session, url, headers=None):
        try:
            async with session.post(url, headers=headers, data={"hello": "post"}) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """ Clear all spans before a test run """
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # New event loop for every test
        # self.loop = tornado.ioloop.IOLoop.current()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self.http_client = AsyncHTTPClient()
        yield
        self.http_client.close()

    def test_get(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_span.p == aiohttp_span.s

        # Synthetic
        assert not tornado_span.sy
        assert not aiohttp_span.sy
        assert not test_span.sy

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not tornado_span.ec

        assert tornado_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == tornado_span.data["http"]["url"]
        assert not tornado_span.data["http"]["params"]
        assert tornado_span.data["http"]["method"] == "GET"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_post(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.post(session, testenv["tornado_server"] + "/")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert not get_current_span().is_recording()

        assert tornado_span.n == "tornado-server"
        assert aiohttp_span.n == "aiohttp-client"
        assert test_span.n == "sdk"

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not tornado_span.ec

        assert tornado_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == tornado_span.data["http"]["url"]
        assert not tornado_span.data["http"]["params"]
        assert tornado_span.data["http"]["method"] == "POST"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "POST"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_synthetic_request(self) -> None:
        async def test():
            headers = {
                'X-INSTANA-SYNTHETIC': '1'
            }

            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/", headers=headers)

        tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span.sy
        assert not aiohttp_span.sy
        assert not test_span.sy

    def test_get_301(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/301")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        filter = lambda span: span.n == "tornado-server" and span.data["http"]["status"] == 301
        tornado_301_span = get_first_span_by_filter(spans, filter)
        filter = lambda span: span.n == "tornado-server" and span.data["http"]["status"] == 200
        tornado_span = get_first_span_by_filter(spans, filter)
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_301_span
        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert not get_current_span().is_recording()

        assert tornado_301_span.n == "tornado-server"
        assert tornado_span.n == "tornado-server"
        assert aiohttp_span.n == "aiohttp-client"
        assert test_span.n == "sdk"

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t
        assert traceId == tornado_301_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_301_span.p == aiohttp_span.s
        assert tornado_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not tornado_301_span.ec
        assert not tornado_span.ec

        assert tornado_301_span.data["http"]["status"] == 301
        assert testenv["tornado_server"] + "/301" == tornado_301_span.data["http"]["url"]
        assert not tornado_span.data["http"]["params"]
        assert tornado_301_span.data["http"]["method"] == "GET"
        assert not tornado_301_span.stack

        assert tornado_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == tornado_span.data["http"]["url"]
        assert tornado_span.data["http"]["method"] == "GET"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/301" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    @pytest.mark.skip("Non deterministic (flaky) testcase")
    def test_get_405(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/405")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert not get_current_span().is_recording()

        assert tornado_span.n == "tornado-server"
        assert aiohttp_span.n == "aiohttp-client"
        assert test_span.n == "sdk"

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not tornado_span.ec

        assert tornado_span.data["http"]["status"] == 405
        assert testenv["tornado_server"] + "/405" == tornado_span.data["http"]["url"]
        assert not tornado_span.data["http"]["params"]
        assert tornado_span.data["http"]["method"] == "GET"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 405
        assert testenv["tornado_server"] + "/405" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    @pytest.mark.skip("Non deterministic (flaky) testcase")
    def test_get_500(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/500")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert not get_current_span().is_recording()

        assert tornado_span.n == "tornado-server"
        assert aiohttp_span.n == "aiohttp-client"
        assert test_span.n == "sdk"

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert aiohttp_span.ec == 1
        assert tornado_span.ec == 1

        assert tornado_span.data["http"]["status"] == 500
        assert testenv["tornado_server"] + "/500" == tornado_span.data["http"]["url"]
        assert not tornado_span.data["http"]["params"]
        assert tornado_span.data["http"]["method"] == "GET"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 500
        assert testenv["tornado_server"] + "/500" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert 'Internal Server Error' == aiohttp_span.data["http"]["error"]
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    @pytest.mark.skip("Non deterministic (flaky) testcase")
    def test_get_504(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/504")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert not get_current_span().is_recording()

        assert tornado_span.n == "tornado-server"
        assert aiohttp_span.n == "aiohttp-client"
        assert test_span.n == "sdk"

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert aiohttp_span.ec == 1
        assert tornado_span.ec == 1

        assert tornado_span.data["http"]["status"] == 504
        assert testenv["tornado_server"] + "/504" == tornado_span.data["http"]["url"]
        assert not tornado_span.data["http"]["params"]
        assert tornado_span.data["http"]["method"] == "GET"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 504
        assert testenv["tornado_server"] + "/504" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert 'Gateway Timeout' == aiohttp_span.data["http"]["error"]
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_get_with_params_to_scrub(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"], params={"secret": "yeah"})

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        assert not get_current_span().is_recording()

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert tornado_span.n == "tornado-server"
        assert aiohttp_span.n == "aiohttp-client"
        assert test_span.n == "sdk"

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not tornado_span.ec

        assert tornado_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == tornado_span.data["http"]["url"]
        assert tornado_span.data["http"]["params"] == "secret=<redacted>"
        assert tornado_span.data["http"]["method"] == "GET"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.data["http"]["params"] == "secret=<redacted>"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_request_header_capture(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    # Hack together a manual custom request headers list
                    agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

                    request_headers = {
                        "X-Capture-This": "this",
                        "X-Capture-That": "that"
                    }

                    return await self.fetch(session, testenv["tornado_server"], headers=request_headers, params={"secret": "iloveyou"})

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert not get_current_span().is_recording()

        assert tornado_span.n == "tornado-server"
        assert aiohttp_span.n == "aiohttp-client"
        assert test_span.n == "sdk"

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not tornado_span.ec

        assert tornado_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == tornado_span.data["http"]["url"]
        assert tornado_span.data["http"]["params"] == "secret=<redacted>"
        assert tornado_span.data["http"]["method"] == "GET"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.data["http"]["params"] == "secret=<redacted>"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

        assert "X-Capture-This" in tornado_span.data["http"]["header"]
        assert tornado_span.data["http"]["header"]["X-Capture-This"] == "this"
        assert "X-Capture-That" in tornado_span.data["http"]["header"]
        assert tornado_span.data["http"]["header"]["X-Capture-That"] == "that"

    def test_response_header_capture(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    # Hack together a manual custom response headers list
                    agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

                    return await self.fetch(session, testenv["tornado_server"] + "/response_headers", params={"secret": "itsasecret"})

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert tornado_span
        assert aiohttp_span
        assert test_span

        assert not get_current_span().is_recording()

        assert tornado_span.n == "tornado-server"
        assert aiohttp_span.n == "aiohttp-client"
        assert test_span.n == "sdk"

        # Same traceId
        traceId = test_span.t
        assert traceId == aiohttp_span.t
        assert traceId == tornado_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert tornado_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not tornado_span.ec

        assert tornado_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/response_headers" == tornado_span.data["http"]["url"]
        assert tornado_span.data["http"]["params"] == "secret=<redacted>"
        assert tornado_span.data["http"]["method"] == "GET"
        assert not tornado_span.stack

        assert aiohttp_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/response_headers" == aiohttp_span.data["http"]["url"]
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.data["http"]["params"] == "secret=<redacted>"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(tornado_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

        assert "X-Capture-This-Too" in tornado_span.data["http"]["header"]
        assert tornado_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in tornado_span.data["http"]["header"]
        assert tornado_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"



# (c) Copyright IBM Corp. 2024

from time import time_ns
from typing import Optional, Sequence, List

from opentelemetry.trace.status import Status, StatusCode
from opentelemetry.util import types

from instana.span_context import SpanContext


class Event:
    def __init__(
        self,
        name: str,
        attributes: types.Attributes = None,
        timestamp: Optional[int] = None,
    ) -> None:
        self._name = name
        self._attributes = attributes
        if timestamp is None:
            self._timestamp = time_ns()
        else:
            self._timestamp = timestamp

    @property
    def name(self) -> str:
        return self._name

    @property
    def timestamp(self) -> int:
        return self._timestamp

    @property
    def attributes(self) -> types.Attributes:
        return self._attributes


class ReadableSpan:
    """
    Provides read-only access to span attributes.

    Users should NOT be creating these objects directly.
    `ReadableSpan`s are created as a direct result from using the tracing pipeline
    via the `Tracer`.
    """

    def __init__(
        self,
        name: str,
        context: SpanContext,
        parent_id: Optional[str] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        attributes: types.Attributes = {},
        events: Sequence[Event] = [],
        status: Optional[Status] = Status(StatusCode.UNSET),
        stack: Optional[List] = None,
    ) -> None:
        self._name = name
        self._context = context
        self._start_time = start_time or time_ns()
        self._end_time = end_time
        self._duration = (
            self._end_time - self._start_time
            if self._start_time and self._end_time
            else None
        )
        self._attributes = attributes if attributes else {}
        self._events = events
        self._parent_id = parent_id
        self._status = status
        self.stack = stack
        self.synthetic = False
        if context.synthetic:
            self.synthetic = True

    @property
    def name(self) -> str:
        return self._name

    @property
    def context(self) -> SpanContext:
        return self._context

    @property
    def start_time(self) -> Optional[int]:
        return self._start_time

    @property
    def end_time(self) -> Optional[int]:
        return self._end_time

    @property
    def duration(self) -> Optional[int]:
        return self._duration

    @property
    def attributes(self) -> types.Attributes:
        return self._attributes

    @property
    def events(self) -> Sequence[Event]:
        return self._events

    @property
    def status(self) -> Status:
        return self._status

    @property
    def parent_id(self) -> int:
        return self._parent_id



# (c) Copyright IBM Corp. 2024

import abc
import enum


class SamplingPolicy(enum.Enum):
    # IsRecording() == False
    # Span will not be recorded and all events and attributes will be dropped.
    # https://opentelemetry.io/docs/specs/otel/trace/api/#isrecording
    DROP = 0
    # IsRecording() == True, but Sampled flag MUST NOT be set.
    RECORD_ONLY = 1
    # IsRecording() == True AND Sampled flag MUST be set.
    RECORD_AND_SAMPLE = 2


class Sampler(abc.ABC):
    """Samplers choose whether the span is recorded or dropped.

    A variety of sampling algorithms are available, and choosing which sampler
    to use and how to configure it is one of the most confusing parts of
    setting up a tracing system.
    """

    @abc.abstractmethod
    def sampled(self) -> bool:
        """
        Returns if a span was dropped (False) or recorded (True).

        Calling a span “sampled” can mean it was “sampled out” (dropped)
        or “sampled in” (recorded).
        """
        pass


class InstanaSampler(Sampler):
    def __init__(self) -> None:
        # Instana never samples.
        self._sampled: SamplingPolicy = SamplingPolicy.DROP

    def sampled(self) -> bool:
        return False if self._sampled == SamplingPolicy.DROP else True



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import datetime
import json
import logging
import os
from typing import Generator
from unittest.mock import Mock

import pytest
import requests
from mock import MagicMock, patch

from instana.agent.host import AnnounceData, HostAgent
from instana.collector.host import HostCollector
from instana.fsm import Discovery, TheMachine
from instana.options import StandardOptions
from instana.recorder import StanRecorder
from instana.singletons import get_agent
from instana.span.span import InstanaSpan
from instana.span_context import SpanContext


class TestHostAgent:
    @pytest.fixture(autouse=True)
    def _resource(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> Generator[None, None, None]:
        self.agent = get_agent()
        self.span_recorder = None
        self.tracer = None
        yield
        caplog.clear()
        variable_names = (
            "AWS_EXECUTION_ENV",
            "INSTANA_EXTRA_HTTP_HEADERS",
            "INSTANA_ENDPOINT_URL",
            "INSTANA_ENDPOINT_PROXY",
            "INSTANA_AGENT_KEY",
            "INSTANA_LOG_LEVEL",
            "INSTANA_SERVICE_NAME",
            "INSTANA_SECRETS",
            "INSTANA_TAGS",
        )

        for variable_name in variable_names:
            if variable_name in os.environ:
                os.environ.pop(variable_name)

    def test_secrets(self) -> None:
        assert hasattr(self.agent.options, "secrets_matcher")
        assert self.agent.options.secrets_matcher == "contains-ignore-case"
        assert hasattr(self.agent.options, "secrets_list")
        assert self.agent.options.secrets_list == ["key", "pass", "secret"]

    def test_options_have_extra_http_headers(self) -> None:
        assert hasattr(self.agent, "options")
        assert hasattr(self.agent.options, "extra_http_headers")

    def test_has_options(self) -> None:
        assert hasattr(self.agent, "options")
        assert isinstance(self.agent.options, StandardOptions)

    def test_agent_default_log_level(self) -> None:
        assert self.agent.options.log_level == logging.WARNING

    def test_agent_instana_debug(self) -> None:
        os.environ["INSTANA_DEBUG"] = "asdf"
        self.agent.options = StandardOptions()
        assert self.agent.options.log_level == logging.DEBUG

    def test_agent_instana_service_name(self) -> None:
        os.environ["INSTANA_SERVICE_NAME"] = "greycake"
        self.agent.options = StandardOptions()
        assert self.agent.options.service_name == "greycake"

    @pytest.mark.original
    @patch.object(requests.Session, "put")
    def test_announce_is_successful(
        self,
        mock_requests_session_put: MagicMock,
    ) -> None:
        test_pid = 4242
        test_process_name = "test_process"
        test_process_args = ["-v", "-d"]
        test_agent_uuid = "83bf1e09-ab16-4203-abf5-34ee0977023a"

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = (
            "{" f'  "pid": {test_pid}, ' f'  "agentUuid": "{test_agent_uuid}"' "}"
        )

        # This mocks the call to self.agent.client.put
        mock_requests_session_put.return_value = mock_response

        d = Discovery(pid=test_pid, name=test_process_name, args=test_process_args)
        payload = self.agent.announce(d)

        assert "pid" in payload
        assert test_pid == payload["pid"]

        assert "agentUuid" in payload
        assert test_agent_uuid == payload["agentUuid"]

    @pytest.mark.original
    @patch.object(requests.Session, "put")
    def test_announce_fails_with_non_200(
        self,
        mock_requests_session_put: MagicMock,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        test_pid = 4242
        test_process_name = "test_process"
        test_process_args = ["-v", "-d"]

        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_response.content = ""
        mock_requests_session_put.return_value = mock_response

        d = Discovery(pid=test_pid, name=test_process_name, args=test_process_args)
        caplog.set_level(logging.DEBUG, logger="instana")
        caplog.clear()
        payload = self.agent.announce(d)
        assert payload is None
        assert len(caplog.messages) == 1
        assert len(caplog.records) == 1
        assert "response status code" in caplog.messages[0]
        assert "is NOT 200" in caplog.messages[0]

    @pytest.mark.original
    @patch.object(requests.Session, "put")
    def test_announce_fails_with_non_json(
        self,
        mock_requests_session_put: MagicMock,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        test_pid = 4242
        test_process_name = "test_process"
        test_process_args = ["-v", "-d"]

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = ""
        mock_requests_session_put.return_value = mock_response

        d = Discovery(pid=test_pid, name=test_process_name, args=test_process_args)
        caplog.set_level(logging.DEBUG, logger="instana")
        caplog.clear()
        payload = self.agent.announce(d)
        assert payload is None
        assert len(caplog.messages) == 1
        assert len(caplog.records) == 1
        assert "response is not JSON" in caplog.messages[0]

    @pytest.mark.original
    @patch.object(requests.Session, "put")
    def test_announce_fails_with_empty_list_json(
        self,
        mock_requests_session_put: MagicMock,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        test_pid = 4242
        test_process_name = "test_process"
        test_process_args = ["-v", "-d"]

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = "[]"
        mock_requests_session_put.return_value = mock_response

        d = Discovery(pid=test_pid, name=test_process_name, args=test_process_args)
        caplog.set_level(logging.DEBUG, logger="instana")
        caplog.clear()
        payload = self.agent.announce(d)
        assert payload is None
        assert len(caplog.messages) == 1
        assert len(caplog.records) == 1
        assert "payload has no fields" in caplog.messages[0]

    @pytest.mark.original
    @patch.object(requests.Session, "put")
    def test_announce_fails_with_missing_pid(
        self,
        mock_requests_session_put: MagicMock,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        test_pid = 4242
        test_process_name = "test_process"
        test_process_args = ["-v", "-d"]
        test_agent_uuid = "83bf1e09-ab16-4203-abf5-34ee0977023a"

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = "{" f'  "agentUuid": "{test_agent_uuid}"' "}"
        mock_requests_session_put.return_value = mock_response

        d = Discovery(pid=test_pid, name=test_process_name, args=test_process_args)
        caplog.clear()
        payload = self.agent.announce(d)
        assert payload is None
        assert len(caplog.messages) == 1
        assert len(caplog.records) == 1
        assert "response payload has no pid" in caplog.messages[0]

    @pytest.mark.original
    @patch.object(requests.Session, "put")
    def test_announce_fails_with_missing_uuid(
        self,
        mock_requests_session_put: MagicMock,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        test_pid = 4242
        test_process_name = "test_process"
        test_process_args = ["-v", "-d"]

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = "{" f'  "pid": {test_pid} ' "}"
        mock_requests_session_put.return_value = mock_response

        d = Discovery(pid=test_pid, name=test_process_name, args=test_process_args)
        caplog.clear()
        payload = self.agent.announce(d)
        assert payload is None
        assert len(caplog.messages) == 1
        assert len(caplog.records) == 1
        assert "response payload has no agentUuid" in caplog.messages[0]

    @pytest.mark.original
    @patch.object(requests.Session, "get")
    def test_agent_connection_attempt(
        self,
        mock_requests_session_get: MagicMock,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_requests_session_get.return_value = mock_response

        host = self.agent.options.agent_host
        port = self.agent.options.agent_port
        msg = f"Instana host agent found on {host}:{port}"

        result = self.agent.is_agent_listening(host, port)

        assert result
        assert msg in caplog.messages[0]

    @pytest.mark.original
    @patch.object(requests.Session, "get")
    def test_agent_connection_attempt_fails_with_404(
        self,
        mock_requests_session_get: MagicMock,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_requests_session_get.return_value = mock_response

        host = self.agent.options.agent_host
        port = self.agent.options.agent_port
        msg = (
            "The attempt to connect to the Instana host agent on "
            f"{host}:{port} has failed with an unexpected status code. "
            f"Expected HTTP 200 but received: {mock_response.status_code}"
        )

        caplog.clear()
        result = self.agent.is_agent_listening(host, port)

        assert not result
        assert msg in caplog.messages[0]

    def test_init(self) -> None:
        with patch(
            "instana.agent.base.BaseAgent.update_log_level"
        ) as mock_update, patch.object(os, "getpid", return_value=12345):
            agent = HostAgent()
            assert not agent.announce_data
            assert not agent.last_seen
            assert not agent.last_fork_check
            assert agent._boot_pid == 12345

            mock_update.assert_called_once()

            assert isinstance(agent.options, StandardOptions)
            assert isinstance(agent.collector, HostCollector)
            assert isinstance(agent.machine, TheMachine)

    def test_start(
        self,
    ) -> None:
        with patch("instana.collector.host.HostCollector.start") as mock_start:
            agent = HostAgent()
            agent.start()
            mock_start.assert_called_once()

    def test_handle_fork(
        self,
    ) -> None:
        with patch.object(HostAgent, "reset") as mock_reset:
            agent = HostAgent()
            agent.handle_fork()
            mock_reset.assert_called_once()

    def test_reset(
        self,
    ) -> None:
        with patch(
            "instana.collector.host.HostCollector.shutdown"
        ) as mock_shutdown, patch("instana.fsm.TheMachine.reset") as mock_reset:
            agent = HostAgent()
            agent.reset()

            assert not agent.last_seen
            assert not agent.announce_data

            mock_shutdown.assert_called_once_with(report_final=False)
            mock_reset.assert_called_once()

    def test_is_timed_out(
        self,
    ) -> None:
        agent = HostAgent()
        assert not agent.is_timed_out()

        agent.last_seen = datetime.datetime.now() - datetime.timedelta(minutes=5)
        agent.can_send = True
        assert agent.is_timed_out()

    def test_can_send_test_env(
        self,
    ) -> None:
        agent = HostAgent()
        with patch.dict("os.environ", {"INSTANA_TEST": "sample-data"}):
            if "INSTANA_TEST" in os.environ:
                assert agent.can_send()

    @pytest.mark.original
    def test_can_send(
        self,
    ) -> None:
        agent = HostAgent()
        agent._boot_pid = 12345
        with patch.object(os, "getpid", return_value=12344), patch(
            "instana.agent.host.HostAgent.handle_fork"
        ) as mock_handle, patch.dict("os.environ", {}, clear=True):
            agent.can_send()
            assert agent._boot_pid == 12344
            mock_handle.assert_called_once()

            with patch.object(agent.machine.fsm, "current", "wait4init"):
                assert agent.can_send() is True

    @pytest.mark.original
    def test_can_send_default(
        self,
    ) -> None:
        agent = HostAgent()
        with patch.dict("os.environ", {}, clear=True):
            assert not agent.can_send()

    def test_set_from(
        self,
    ) -> None:
        agent = HostAgent()
        sample_res_data = {
            "secrets": {"matcher": "value-1", "list": ["value-2"]},
            "extraHeaders": ["value-3"],
            "agentUuid": "value-4",
            "pid": 1234,
        }
        agent.options.extra_http_headers = None

        agent.set_from(sample_res_data)
        assert agent.options.secrets_matcher == "value-1"
        assert agent.options.secrets_list == ["value-2"]
        assert agent.options.extra_http_headers == ["value-3"]

        agent.options.extra_http_headers = ["value"]
        agent.set_from(sample_res_data)
        assert "value" in agent.options.extra_http_headers

        assert agent.announce_data.agentUuid == "value-4"
        assert agent.announce_data.pid == 1234

    @pytest.mark.original
    def test_get_from_structure(
        self,
    ) -> None:
        agent = HostAgent()
        agent.announce_data = AnnounceData(pid=1234, agentUuid="value")
        assert agent.get_from_structure() == {"e": 1234, "h": "value"}

    @pytest.mark.original
    def test_is_agent_listening(
        self,
        caplog,
    ) -> None:
        agent = HostAgent()
        mock_response = Mock()
        mock_response.status_code = 200
        with patch.object(requests.Session, "get", return_value=mock_response):
            assert agent.is_agent_listening("sample", 1234)

        mock_response.status_code = 404
        with patch.object(
            requests.Session, "get", return_value=mock_response, clear=True
        ):
            assert not agent.is_agent_listening("sample", 1234)

        host = "localhost"
        port = 123
        with patch.object(requests.Session, "get", side_effect=Exception()):
            caplog.set_level(logging.DEBUG, logger="instana")
            agent.is_agent_listening(host, port)
            assert f"Instana Host Agent not found on {host}:{port}" in caplog.messages

    @pytest.mark.original
    def test_announce(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        agent = HostAgent()
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.content = json.dumps(
            {"get": "value", "pid": "value", "agentUuid": "value"}
        )
        response = json.loads(mock_response.content)
        with patch.object(requests.Session, "put", return_value=mock_response):
            assert agent.announce("sample-data") == response

        mock_response.content = mock_response.content.encode("UTF-8")
        with patch.object(requests.Session, "put", return_value=mock_response):
            assert agent.announce("sample-data") == response

        mock_response.content = json.dumps(
            {"get": "value", "pid": "value", "agentUuid": "value"}
        )

        with patch.object(requests.Session, "put", side_effect=Exception()):
            caplog.set_level(logging.DEBUG, logger="instana")
            assert not agent.announce("sample-data")
            assert (
                f"announce: connection error ({type(Exception())})" in caplog.messages
            )

        mock_response.content = json.dumps("key")
        with patch.object(
            requests.Session, "put", return_value=mock_response, clear=True
        ):
            caplog.set_level(logging.DEBUG, logger="instana")
            assert not agent.announce("sample-data")
            assert "announce: response payload has no fields: (key)" in caplog.messages

        mock_response.content = json.dumps({"key": "value"})
        with patch.object(
            requests.Session, "put", return_value=mock_response, clear=True
        ):
            caplog.set_level(logging.DEBUG, logger="instana")
            assert not agent.announce("sample-data")
            assert (
                "announce: response payload has no pid: ({'key': 'value'})"
                in caplog.messages
            )

        mock_response.content = json.dumps({"pid": "value"})
        with patch.object(
            requests.Session, "put", return_value=mock_response, clear=True
        ):
            caplog.set_level(logging.DEBUG, logger="instana")
            assert not agent.announce("sample-data")
            assert (
                "announce: response payload has no agentUuid: ({'pid': 'value'})"
                in caplog.messages
            )

        mock_response.status_code = 404
        with patch.object(
            requests.Session, "put", return_value=mock_response, clear=True
        ):
            assert not agent.announce("sample-data")
            assert "announce: response status code (404) is NOT 200" in caplog.messages

    def test_log_message_to_host_agent(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        agent = HostAgent()
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.return_value = "sample"
        mock_datetime = datetime.datetime(2022, 1, 1, 12, 0, 0)
        with patch.object(requests.Session, "post", return_value=mock_response), patch(
            "instana.agent.host.datetime"
        ) as mock_date:
            mock_date.now.return_value = mock_datetime
            mock_date.side_effect = lambda *args, **kwargs: datetime(*args, **kwargs)
            agent.log_message_to_host_agent("sample")
            assert agent.last_seen == mock_datetime

            with patch.object(requests.Session, "post", side_effect=Exception()):
                caplog.set_level(logging.DEBUG, logger="instana")
                agent.log_message_to_host_agent("sample")
                assert (
                    f"agent logging: connection error ({type(Exception())})"
                    in caplog.messages
                )

    def test_is_agent_ready(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        agent = HostAgent()
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.return_value = {"key": "value"}
        agent.AGENT_DATA_PATH = "sample_path"
        agent.announce_data = AnnounceData(pid=1234, agentUuid="sample")
        with patch.object(requests.Session, "head", return_value=mock_response), patch(
            "instana.agent.host.HostAgent._HostAgent__data_url",
            return_value="localhost",
        ):
            assert agent.is_agent_ready()
            with patch.object(requests.Session, "head", side_effect=Exception()):
                caplog.set_level(logging.DEBUG, logger="instana")
                agent.is_agent_ready()
                assert (
                    f"is_agent_ready: connection error ({type(Exception())})"
                    in caplog.messages
                )

    def test_report_data_payload(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        agent = HostAgent()
        span_name = "test-span"
        span_1 = InstanaSpan(span_name, span_context, span_processor)
        span_2 = InstanaSpan(span_name, span_context, span_processor)
        payload = {
            "spans": [span_1, span_2],
            "profiles": ["profile-1", "profile-2"],
            "metrics": {
                "plugins": [
                    {"data": "sample data"},
                ]
            },
        }
        sample_response = {"key": "value"}
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.content = sample_response
        with patch.object(requests.Session, "post", return_value=mock_response), patch(
            "instana.agent.host.HostAgent._HostAgent__traces_url",
            return_value="localhost",
        ), patch(
            "instana.agent.host.HostAgent._HostAgent__profiles_url",
            return_value="localhost",
        ), patch(
            "instana.agent.host.HostAgent._HostAgent__data_url",
            return_value="localhost",
        ):
            test_response = agent.report_data_payload(payload)
            assert isinstance(agent.last_seen, datetime.datetime)
            assert test_response.content == sample_response

    def test_diagnostics(self, caplog: pytest.LogCaptureFixture) -> None:
        caplog.set_level(logging.WARNING, logger="instana")

        agent = HostAgent()
        agent.diagnostics()
        assert (
            "====> Instana Python Language Agent Diagnostics <====" in caplog.messages
        )
        assert "----> Agent <----" in caplog.messages
        assert f"is_agent_ready: {agent.is_agent_ready()}" in caplog.messages
        assert f"is_timed_out: {agent.is_timed_out()}" in caplog.messages
        assert "last_seen: None" in caplog.messages

        sample_date = datetime.datetime(2022, 7, 25, 14, 30, 0)
        agent.last_seen = sample_date
        agent.diagnostics()
        assert "last_seen: 2022-07-25 14:30:00" in caplog.messages
        assert "announce_data: None" in caplog.messages

        agent.announce_data = AnnounceData(pid=1234, agentUuid="value")
        agent.diagnostics()
        assert f"announce_data: {agent.announce_data.__dict__}" in caplog.messages
        assert f"Options: {agent.options.__dict__}" in caplog.messages
        assert "----> StateMachine <----" in caplog.messages
        assert f"State: {agent.machine.fsm.current}" in caplog.messages
        assert "----> Collector <----" in caplog.messages
        assert f"Collector: {agent.collector}" in caplog.messages
        assert f"ready_to_start: {agent.collector.ready_to_start}" in caplog.messages
        assert "reporting_thread: None" in caplog.messages
        assert f"report_interval: {agent.collector.report_interval}" in caplog.messages
        assert "should_send_snapshot_data: True" in caplog.messages



# (c) Copyright IBM Corp. 2024

import logging
import time
from typing import Any, Dict, Generator, Tuple

import pytest
from opentelemetry.trace import SpanKind

from instana.recorder import StanRecorder
from instana.span.registered_span import RegisteredSpan
from instana.span.span import InstanaSpan
from instana.span_context import SpanContext


class TestRegisteredSpan:
    @pytest.fixture(autouse=True)
    def _resource(
        self,
    ) -> Generator[None, None, None]:
        self.span = None
        yield

    @pytest.mark.parametrize(
        "span_name, expected_result, attributes",
        [
            ("wsgi", ("wsgi", SpanKind.SERVER, "http"), {}),
            ("rabbitmq", ("rabbitmq", SpanKind.SERVER, "rabbitmq"), {}),
            ("gcps-producer", ("gcps", SpanKind.CLIENT, "gcps"), {}),
            ("urllib3", ("urllib3", SpanKind.CLIENT, "http"), {}),
            (
                "rabbitmq",
                ("rabbitmq", SpanKind.CLIENT, "rabbitmq"),
                {"sort": "publish"},
            ),
            (
                "render",
                ("render", SpanKind.INTERNAL, "render"),
                {"arguments": "--quiet"},
            ),
        ],
    )
    def test_registered_span(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        span_name: str,
        expected_result: Tuple[str, int, str],
        attributes: Dict[str, Any],
    ) -> None:
        service_name = "test-registered-service"
        self.span = InstanaSpan(
            span_name, span_context, span_processor, attributes=attributes
        )
        reg_span = RegisteredSpan(self.span, None, service_name)

        assert expected_result[0] == reg_span.n
        assert expected_result[1] == reg_span.k
        assert service_name == reg_span.data["service"]
        assert expected_result[2] in reg_span.data.keys()

    def test_collect_http_attributes_with_attributes(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-registered-span"
        attributes = {
            "span.kind": "entry",
            "http.host": "localhost",
            "http.url": "https://www.instana.com",
            "http.header.test": "one more test",
        }
        service_name = "test-registered-service"
        self.span = InstanaSpan(
            span_name, span_context, span_processor, attributes=attributes
        )
        reg_span = RegisteredSpan(self.span, None, service_name)

        excepted_result = {
            "http.host": attributes["http.host"],
            "http.url": attributes["http.url"],
            "http.header.test": attributes["http.header.test"],
        }

        reg_span._collect_http_attributes(self.span)

        assert excepted_result["http.host"] == reg_span.data["http"]["host"]
        assert excepted_result["http.url"] == reg_span.data["http"]["url"]
        assert (
            excepted_result["http.header.test"]
            == reg_span.data["http"]["header"]["test"]
        )

    def test_populate_local_span_data_with_other_name(
        self,
        span_context: SpanContext,
        span_processor,
        caplog,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        span_name = "test-registered-span"
        service_name = "test-registered-service"
        self.span = InstanaSpan(span_name, span_context, span_processor)
        reg_span = RegisteredSpan(self.span, None, service_name)

        expected_msg = f"SpanRecorder: Unknown local span: {span_name}"

        reg_span._populate_local_span_data(self.span)

        assert expected_msg in caplog.messages

    @pytest.mark.parametrize(
        "span_name, service_name, attributes",
        [
            (
                "aws.lambda.entry",
                "lambda",
                {
                    "lambda.arn": "test",
                    "lambda.trigger": None,
                },
            ),
            (
                "celery-worker",
                "celery",
                {
                    "host": "localhost",
                    "port": 1234,
                },
            ),
            (
                "gcps-consumer",
                "gcps",
                {
                    "gcps.op": "consume",
                    "gcps.projid": "MY_PROJECT",
                    "gcps.sub": "MY_SUBSCRIPTION_NAME",
                },
            ),
            (
                "rpc-server",
                "rpc",
                {
                    "rpc.flavor": "Vanilla",
                    "rpc.host": "localhost",
                    "rpc.port": 1234,
                },
            ),
        ],
    )
    def test_populate_entry_span_data(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        span_name: str,
        service_name: str,
        attributes: Dict[str, Any],
    ) -> None:
        self.span = InstanaSpan(span_name, span_context, span_processor)
        reg_span = RegisteredSpan(self.span, None, service_name)

        expected_result = {}
        for attr, value in attributes.items():
            attrl = attr.split(".")
            attrl = attrl[1] if len(attrl) > 1 else attrl[0]
            expected_result[attrl] = value

        self.span.set_attributes(attributes)
        reg_span._populate_entry_span_data(self.span)

        for attr, value in expected_result.items():
            assert value == reg_span.data[service_name][attr]

    @pytest.mark.parametrize(
        "attributes",
        [
            {
                "lambda.arn": "test",
                "lambda.trigger": "aws:api.gateway",
                "http.host": "localhost",
                "http.url": "https://www.instana.com",
            },
            {
                "lambda.arn": "test",
                "lambda.trigger": "aws:cloudwatch.events",
                "lambda.cw.events.resources": "Resource 1",
            },
            {
                "lambda.arn": "test",
                "lambda.trigger": "aws:cloudwatch.logs",
                "lambda.cw.logs.group": "My Group",
            },
            {
                "lambda.arn": "test",
                "lambda.trigger": "aws:s3",
                "lambda.s3.events": "Event 1",
            },
            {
                "lambda.arn": "test",
                "lambda.trigger": "aws:sqs",
                "lambda.sqs.messages": "Message 1",
            },
        ],
    )
    def test_populate_entry_span_data_AWSlambda(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        attributes: Dict[str, Any],
    ) -> None:
        span_name = "aws.lambda.entry"
        service_name = "lambda"
        expected_result = attributes.copy()

        self.span = InstanaSpan(span_name, span_context, span_processor)
        reg_span = RegisteredSpan(self.span, None, service_name)

        self.span.set_attributes(attributes)
        reg_span._populate_entry_span_data(self.span)

        assert "python" == reg_span.data["lambda"]["runtime"]
        assert "Unknown" == reg_span.data["lambda"]["functionName"]
        assert "test" == reg_span.data["lambda"]["arn"]
        assert expected_result["lambda.trigger"] == reg_span.data["lambda"]["trigger"]

        if expected_result["lambda.trigger"] == "aws:api.gateway":
            assert expected_result["http.host"] == reg_span.data["http"]["host"]
            assert expected_result["http.url"] == reg_span.data["http"]["url"]

        elif expected_result["lambda.trigger"] == "aws:cloudwatch.events":
            assert (
                expected_result["lambda.cw.events.resources"]
                == reg_span.data["lambda"]["cw"]["events"]["resources"]
            )
        elif expected_result["lambda.trigger"] == "aws:cloudwatch.logs":
            assert (
                expected_result["lambda.cw.logs.group"]
                == reg_span.data["lambda"]["cw"]["logs"]["group"]
            )
        elif expected_result["lambda.trigger"] == "aws:s3":
            assert (
                expected_result["lambda.s3.events"]
                == reg_span.data["lambda"]["s3"]["events"]
            )
        elif expected_result["lambda.trigger"] == "aws:sqs":
            assert (
                expected_result["lambda.sqs.messages"]
                == reg_span.data["lambda"]["sqs"]["messages"]
            )

    @pytest.mark.parametrize(
        "span_name, service_name, attributes",
        [
            (
                "cassandra",
                "cassandra",
                {
                    "cassandra.cluster": "my_cluster",
                    "cassandra.error": "minor error",
                },
            ),
            (
                "celery-client",
                "celery",
                {
                    "host": "localhost",
                    "port": 1234,
                },
            ),
            (
                "couchbase",
                "couchbase",
                {
                    "couchbase.hostname": "localhost",
                    "couchbase.error_type": 1234,
                },
            ),
            (
                "rabbitmq",
                "rabbitmq",
                {
                    "address": "localhost",
                    "key": 1234,
                },
            ),
            (
                "redis",
                "redis",
                {
                    "command": "ls -l",
                    "redis.error": "minor error",
                },
            ),
            (
                "rpc-client",
                "rpc",
                {
                    "rpc.flavor": "Vanilla",
                    "rpc.host": "localhost",
                    "rpc.port": 1234,
                },
            ),
            (
                "sqlalchemy",
                "sqlalchemy",
                {
                    "sqlalchemy.sql": "SELECT * FROM everything;",
                    "sqlalchemy.err": "Impossible select everything from everything!",
                },
            ),
            (
                "mysql",
                "mysql",
                {
                    "host": "localhost",
                    "port": 1234,
                },
            ),
            (
                "postgres",
                "pg",
                {
                    "host": "localhost",
                    "port": 1234,
                },
            ),
            (
                "mongo",
                "mongo",
                {
                    "command": "IDK",
                    "error": "minor error",
                },
            ),
            (
                "gcs",
                "gcs",
                {
                    "gcs.op": "produce",
                    "gcs.projectId": "MY_PROJECT",
                    "gcs.accessId": "Can not tell you!",
                },
            ),
            (
                "gcps-producer",
                "gcps",
                {
                    "gcps.op": "produce",
                    "gcps.projid": "MY_PROJECT",
                    "gcps.top": "MY_SUBSCRIPTION_NAME",
                },
            ),
        ],
    )
    def test_populate_exit_span_data(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        span_name: str,
        service_name: str,
        attributes: Dict[str, Any],
    ) -> None:
        self.span = InstanaSpan(span_name, span_context, span_processor)
        reg_span = RegisteredSpan(self.span, None, service_name)

        expected_result = {}
        for attr, value in attributes.items():
            attrl = attr.split(".")
            attrl = attrl[1] if len(attrl) > 1 else attrl[0]
            expected_result[attrl] = value

        self.span.set_attributes(attributes)
        reg_span._populate_exit_span_data(self.span)

        for attr, value in expected_result.items():
            assert value == reg_span.data[service_name][attr]

    @pytest.mark.parametrize(
        "attributes",
        [
            {
                "op": "test",
                "http.host": "localhost",
                "http.url": "https://www.instana.com",
            },
            {
                "payload": {
                    "blah": "bleh",
                    "blih": "bloh",
                },
                "http.host": "localhost",
                "http.url": "https://www.instana.com",
            },
        ],
    )
    def test_populate_exit_span_data_boto3(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        attributes: Dict[str, Any],
    ) -> None:
        span_name = service_name = "boto3"
        expected_result = attributes.copy()

        self.span = InstanaSpan(span_name, span_context, span_processor)
        reg_span = RegisteredSpan(self.span, None, service_name)

        self.span.set_attributes(attributes)
        reg_span._populate_exit_span_data(self.span)

        assert expected_result.pop("http.host", None) == reg_span.data["http"]["host"]
        assert expected_result.pop("http.url", None) == reg_span.data["http"]["url"]

        for attr, value in expected_result.items():
            assert value == reg_span.data[service_name][attr]

    def test_populate_exit_span_data_log(
        self, span_context: SpanContext, span_processor: StanRecorder
    ) -> None:
        span_name = service_name = "log"
        self.span = InstanaSpan(span_name, span_context, span_processor)
        reg_span = RegisteredSpan(self.span, None, service_name)

        excepted_text = "Houston, we have a problem!"
        sample_events = [
            (
                "test_populate_exit_span_data_log_event_with_message",
                {
                    "field1": 1,
                    "field2": "two",
                    "message": excepted_text,
                },
                time.time_ns(),
            ),
            (
                "test_populate_exit_span_data_log_event_with_parameters",
                {
                    "field1": 1,
                    "field2": "two",
                    "parameters": excepted_text,
                },
                time.time_ns(),
            ),
        ]

        for event_name, attributes, timestamp in sample_events:
            self.span.add_event(event_name, attributes, timestamp)

        reg_span._populate_exit_span_data(self.span)

        assert excepted_text == reg_span.data["log"]["message"]
        assert excepted_text == reg_span.data["log"]["parameters"]

        while self.span._events:
            self.span._events.pop()



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
cassandra instrumentation
https://docs.datastax.com/en/developer/python-driver/3.20/
https://github.com/datastax/python-driver
"""

from typing import Any, Callable, Dict, Tuple
import wrapt
from instana.log import logger
from instana.span.span import InstanaSpan
from instana.util.traceutils import get_tracer_tuple, tracing_is_off

try:
    import cassandra
    from cassandra.cluster import ResponseFuture, Session

    consistency_levels = dict(
        {
            0: "ANY",
            1: "ONE",
            2: "TWO",
            3: "THREE",
            4: "QUORUM",
            5: "ALL",
            6: "LOCAL_QUORUM",
            7: "EACH_QUORUM",
            8: "SERIAL",
            9: "LOCAL_SERIAL",
            10: "LOCAL_ONE",
        }
    )

    def collect_attributes(
        span: InstanaSpan,
        fn: ResponseFuture,
    ) -> None:
        tried_hosts = []
        for host in fn.attempted_hosts:
            tried_hosts.append(f"{host.endpoint.address}:{host.endpoint.port}")

        span.set_attribute("cassandra.triedHosts", tried_hosts)
        span.set_attribute("cassandra.coordHost", fn.coordinator_host)

        cl = fn.query.consistency_level
        if cl and cl in consistency_levels:
            span.set_attribute("cassandra.achievedConsistency", consistency_levels[cl])

    def cb_request_finish(
        _,
        span: InstanaSpan,
        fn: ResponseFuture,
    ) -> None:
        collect_attributes(span, fn)
        span.end()

    def cb_request_error(
        results: Dict[str, Any],
        span: InstanaSpan,
        fn: ResponseFuture,
    ) -> None:
        collect_attributes(span, fn)
        span.mark_as_errored({"cassandra.error": results.summary})
        span.end()

    def request_init_with_instana(
        fn: ResponseFuture,
    ) -> None:
        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        if tracing_is_off():
            return

        attributes = {}
        if isinstance(fn.query, cassandra.query.SimpleStatement):
            attributes["cassandra.query"] = fn.query.query_string
        elif isinstance(fn.query, cassandra.query.BoundStatement):
            attributes["cassandra.query"] = fn.query.prepared_statement.query_string

        attributes["cassandra.keyspace"] = fn.session.keyspace
        attributes["cassandra.cluster"] = fn.session.cluster.metadata.cluster_name

        with tracer.start_as_current_span(
            "cassandra",
            span_context=parent_context,
            attributes=attributes,
            end_on_exit=False,
        ) as span:
            fn.add_callback(cb_request_finish, span, fn)
            fn.add_errback(cb_request_error, span, fn)

    @wrapt.patch_function_wrapper("cassandra.cluster", "Session.__init__")
    def init_with_instana(
        wrapped: Callable[..., object],
        instance: Session,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        session = wrapped(*args, **kwargs)
        instance.add_request_init_listener(request_init_with_instana)
        return session

    logger.debug("Instrumenting cassandra")

except ImportError:
    pass
